# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("qwave.dll")
prototypes = \
    {
        #
        'QOSCreateHandle': SimTypeFunction([SimTypePointer(SimTypeRef("QOS_VERSION", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Version", "QOSHandle"]),
        #
        'QOSCloseHandle': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle"]),
        #
        'QOSStartTrackingClient': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("SOCKADDR", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "DestAddr", "Flags"]),
        #
        'QOSStopTrackingClient': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("SOCKADDR", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "DestAddr", "Flags"]),
        #
        'QOSEnumerateFlows': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "Size", "Buffer"]),
        #
        'QOSAddSocketToFlow': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeRef("SOCKADDR", SimStruct), offset=0), SimTypeInt(signed=False, label="QOS_TRAFFIC_TYPE"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "Socket", "DestAddr", "TrafficType", "Flags", "FlowId"]),
        #
        'QOSRemoveSocketFromFlow': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "Socket", "FlowId", "Flags"]),
        #
        'QOSSetFlow': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="QOS_SET_FLOW"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("OVERLAPPED", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "FlowId", "Operation", "Size", "Buffer", "Flags", "Overlapped"]),
        #
        'QOSQueryFlow': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="QOS_QUERY_FLOW"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("OVERLAPPED", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "FlowId", "Operation", "Size", "Buffer", "Flags", "Overlapped"]),
        #
        'QOSNotifyFlow': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="QOS_NOTIFY_FLOW"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("OVERLAPPED", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "FlowId", "Operation", "Size", "Buffer", "Flags", "Overlapped"]),
        #
        'QOSCancel': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("OVERLAPPED", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "Overlapped"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("netsh.dll")
prototypes = \
    {
        #
        'MatchEnumTag': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("TOKEN_VALUE", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hModule", "pwcArg", "dwNumArg", "pEnumTable", "pdwValue"]),
        #
        'MatchToken': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwszUserToken", "pwszCmdToken"]),
        #
        'PreprocessCommand': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("TAG_TYPE", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hModule", "ppwcArguments", "dwCurrentIndex", "dwArgCount", "pttTags", "dwTagCount", "dwMinArgs", "dwMaxArgs", "pdwTagType"]),
        #
        'PrintError': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hModule", "dwErrId"]),
        #
        'PrintMessageFromModule': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hModule", "dwMsgId"]),
        #
        'PrintMessage': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pwszFormat"]),
        #
        'RegisterContext': SimTypeFunction([SimTypePointer(SimTypeRef("NS_CONTEXT_ATTRIBUTES", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pChildContext"]),
        #
        'RegisterHelper': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeRef("NS_HELPER_ATTRIBUTES", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pguidParentContext", "pfnRegisterSubContext"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations
from .base import SimSootValue


class SimSootValue_ParamRef(SimSootValue):
    __slots__ = ["id", "index", "type"]

    def __init__(self, index, type_):
        self.id = f"param_{index}"
        self.index = index
        self.type = type_

    def __repr__(self):
        return self.id

    @classmethod
    def from_sootvalue(cls, soot_value, state):
        return cls(soot_value.index, soot_value.type)



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
from angr.ailment.expression import BinaryOp, Const, Convert

from .base import PeepholeOptimizationExprBase


class ModuloSimplifier(PeepholeOptimizationExprBase):
    """
    Simplify division and multiplication expressions that can be reduced to a modulo operation.
    """

    __slots__ = ()

    NAME = "a - (a / N) * N => a % N"
    expr_classes = (BinaryOp,)

    def optimize(  # pylint:disable=unused-argument
        self, expr: BinaryOp, stmt_idx: int | None = None, block=None, **kwargs
    ):
        if expr.op == "Sub" and len(expr.operands) == 2:
            sub0, sub1 = expr.operands
            # unpack Conversions
            outer_conv_expr = None
            if (
                isinstance(sub0, Convert)
                and isinstance(sub1, Convert)
                and sub0.to_bits == sub1.to_bits
                and sub0.from_bits == sub1.from_bits
                and sub0.to_bits > sub0.from_bits
                and sub0.is_signed == sub1.is_signed
            ):
                # Convert(a) - Convert(a / N * N)  ==> Convert(a % N)
                outer_conv_expr = sub0
                sub0 = sub0.operand
                sub1 = sub1.operand

            if isinstance(sub1, BinaryOp) and sub1.op == "Mul" and isinstance(sub1.operands[1], Const):
                a0, op1 = sub0, sub1
                op1_left = op1.operands[0]
                mul_const = sub1.operands[1]

                if (
                    isinstance(op1_left, Convert)
                    and isinstance(a0, Convert)
                    and op1_left.to_bits == a0.to_bits
                    and op1_left.from_bits == a0.from_bits
                ):
                    # Convert(a) - (Convert(a / N)) * N  ==>  Convert(a) % N
                    inner_conv_expr = a0
                    a0 = a0.operand
                    op1_left = op1_left.operand
                else:
                    inner_conv_expr = None

                if isinstance(op1_left, BinaryOp) and op1_left.op == "Div" and isinstance(op1_left.operands[1], Const):
                    # a - (a / N) * N  ==>  a % N
                    a1 = op1_left.operands[0]
                    div_const = op1_left.operands[1]

                    if a0.likes(a1) and mul_const.value == div_const.value:
                        operands = [a0, div_const]
                        mod = BinaryOp(expr.idx, "Mod", operands, False, bits=a0.bits, **expr.tags)
                        if inner_conv_expr is not None:
                            conv_from_bits = inner_conv_expr.from_bits
                            conv_to_bits = (
                                inner_conv_expr.to_bits if outer_conv_expr is None else outer_conv_expr.to_bits
                            )
                            conv_signed = inner_conv_expr.is_signed
                            conv_expr = inner_conv_expr
                        elif outer_conv_expr is not None:
                            conv_from_bits = outer_conv_expr.from_bits
                            conv_to_bits = outer_conv_expr.to_bits
                            conv_signed = outer_conv_expr.is_signed
                            conv_expr = outer_conv_expr
                        else:
                            # no conversion necessary
                            return mod

                        return Convert(
                            conv_expr.idx,
                            conv_from_bits,
                            conv_to_bits,
                            conv_signed,
                            mod,
                            **conv_expr.tags,
                        )

        return None



# pylint:disable=missing-class-docstring
"""
All type constants used in type inference. They can be mapped, translated, or rewritten to C-style types.
"""
from __future__ import annotations

import functools


def memoize(f):
    @functools.wraps(f)
    def wrapped_repr(self, *args, **kwargs):
        memo = set() if not kwargs or "memo" not in kwargs else kwargs.pop("memo")
        if self in memo:
            return "..."
        memo.add(self)
        r = f(self, *args, memo=memo, **kwargs)
        memo.remove(self)
        return r

    return wrapped_repr


class TypeConstant:
    SIZE = None

    def pp_str(self, mapping) -> str:  # pylint:disable=unused-argument
        return repr(self)

    def _hash(self, visited: set[int]):  # pylint:disable=unused-argument
        return hash(type(self))

    def __eq__(self, other):
        return type(self) is type(other)

    def __hash__(self):
        return self._hash(set())

    @property
    def size(self) -> int:
        if self.SIZE is None:
            raise NotImplementedError
        return self.SIZE

    def __repr__(self, memo=None) -> str:
        raise NotImplementedError


class TopType(TypeConstant):
    def __repr__(self, memo=None):
        return "TOP"


class BottomType(TypeConstant):
    def __repr__(self, memo=None):
        return "BOT"


class Int(TypeConstant):
    def __repr__(self, memo=None) -> str:
        return "intbase"


class Int1(Int):
    SIZE = 1


class Int8(Int):
    SIZE = 1

    def __repr__(self, memo=None):
        return "int8"


class Int16(Int):
    SIZE = 2

    def __repr__(self, memo=None):
        return "int16"


class Int32(Int):
    SIZE = 4

    def __repr__(self, memo=None) -> str:
        return "int32"


class Int64(Int):
    SIZE = 8

    def __repr__(self, memo=None) -> str:
        return "int64"


class Int128(Int):
    SIZE = 16

    def __repr__(self, memo=None):
        return "int128"


class Int256(Int):
    SIZE = 32

    def __repr__(self, memo=None):
        return "int256"


class Int512(Int):
    SIZE = 32

    def __repr__(self, memo=None):
        return "int512"


class IntVar(Int):
    def __init__(self, size):
        self._size = size

    @property
    def size(self) -> int:
        return self._size

    def __repr__(self, memo=None):
        return "intvar"


class Float(TypeConstant):
    def __repr__(self, memo=None) -> str:
        return "floatbase"


class Float32(Float):
    SIZE = 4

    def __repr__(self, memo=None):
        return "float32"


class Float64(Float):
    SIZE = 8

    def __repr__(self, memo=None):
        return "float64"


class Pointer(TypeConstant):
    def __init__(self, basetype: TypeConstant | None):
        self.basetype: TypeConstant | None = basetype

    def __eq__(self, other):
        return type(self) is type(other) and self.basetype == other.basetype

    def _hash(self, visited: set[int]):
        if self.basetype is None:
            return hash(type(self))
        return hash((type(self), self.basetype._hash(visited)))

    def new(self, basetype):
        return self.__class__(basetype)

    def __hash__(self):
        return self._hash(set())


class Pointer32(Pointer, Int32):
    """
    32-bit pointers.
    """

    def __init__(self, basetype=None):
        Pointer.__init__(self, basetype)

    @memoize
    def __repr__(self, memo=None):
        bt = self.basetype.__repr__(memo=memo) if isinstance(self.basetype, TypeConstant) else repr(self.basetype)
        return f"ptr32({bt})"


class Pointer64(Pointer, Int64):
    """
    64-bit pointers.
    """

    def __init__(self, basetype=None):
        Pointer.__init__(self, basetype)

    @memoize
    def __repr__(self, memo=None):
        bt = self.basetype.__repr__(memo=memo) if isinstance(self.basetype, TypeConstant) else repr(self.basetype)
        return f"ptr64({bt})"


class Array(TypeConstant):
    def __init__(self, element=None, count=None):
        self.element: TypeConstant | None = element
        self.count: int | None = count

    @property
    def size(self) -> int:
        if not self.count or not self.element:
            return 0
        return self.element.size * self.count

    @memoize
    def __repr__(self, memo=None):
        if self.count is None:
            return f"{self.element!r}[?]"
        return f"{self.element!r}[{self.count}]"

    def __eq__(self, other):
        return type(other) is type(self) and self.element == other.element and self.count == other.count

    def _hash(self, visited: set[int]):
        if id(self) in visited:
            return 0
        visited.add(id(self))
        return hash((type(self), self.element, self.count))

    def __hash__(self):
        return self._hash(set())


class Struct(TypeConstant):
    def __init__(self, fields=None, name=None, field_names=None, is_cppclass: bool = False):
        self.fields = {} if fields is None else fields  # offset to type
        self.name = name
        self.field_names = field_names
        self.is_cppclass = is_cppclass

    def _hash(self, visited: set[int]):
        if id(self) in visited:
            return 0
        visited.add(id(self))
        return hash((type(self), self._hash_fields(visited)))

    def _hash_fields(self, visited: set[int]):
        keys = sorted(self.fields.keys())
        tpl = tuple((k, self.fields[k]._hash(visited) if self.fields[k] is not None else None) for k in keys)
        return hash(tpl)

    @property
    def size(self) -> int:
        if not self.fields:
            return 0
        max_field_off = max(self.fields.keys())
        return max_field_off + (
            self.fields[max_field_off].size if not isinstance(self.fields[max_field_off], BottomType) else 1
        )

    @memoize
    def __repr__(self, memo=None):
        prefix = "CppClass" if self.is_cppclass else "struct"
        if self.name:
            prefix = f"{prefix} {self.name}"
        return (
            prefix
            + "{"
            + ", ".join(f"{k}:{v.__repr__(memo=memo) if v is not None else 'None'}" for k, v in self.fields.items())
            + "}"
        )

    def __eq__(self, other):
        return type(other) is type(self) and hash(self) == hash(other)

    def __hash__(self):
        return self._hash(set())


class Function(TypeConstant):
    def __init__(self, params: list, outputs: list):
        self.params = params
        self.outputs = outputs

    @memoize
    def __repr__(self, memo=None):
        param_str = ", ".join(repr(param) for param in self.params)
        outputs_str = ", ".join(repr(output) for output in self.outputs)
        return f"func({param_str}) -> {outputs_str}"

    def __eq__(self, other):
        if not isinstance(other, Function):
            return False
        return self.params == other.params and self.outputs == other.outputs

    def _hash(self, visited: set[int]):
        if id(self) in visited:
            return 0
        visited.add(id(self))

        params_hash = tuple(param._hash(visited) for param in self.params)
        outputs_hash = tuple(out._hash(visited) for out in self.outputs)
        return hash((Function, params_hash, outputs_hash))

    def __hash__(self):
        return self._hash(set())


class TypeVariableReference(TypeConstant):
    def __init__(self, typevar):
        self.typevar = typevar

    def __repr__(self, memo=None):
        return f"ref({self.typevar})"

    def __eq__(self, other):
        return type(other) is type(self) and self.typevar == other.typevar

    def __hash__(self):
        return hash((type(self), self.typevar))


#
# Methods
#


def int_type(bits: int) -> Int:
    mapping = {
        1: Int1,
        8: Int8,
        16: Int16,
        32: Int32,
        64: Int64,
        128: Int128,
        256: Int256,
        512: Int512,
    }
    return mapping[bits]() if bits in mapping else IntVar(bits)


def float_type(bits: int) -> Float | None:
    if bits == 32:
        return Float32()
    if bits == 64:
        return Float64()
    return None



# pylint:disable=arguments-renamed,global-statement
from __future__ import annotations
import copy
import os
import logging
import inspect
from collections import defaultdict
from typing import TYPE_CHECKING

import pydemumble
import archinfo

from angr.errors import AngrMissingTypeError
from angr.sim_type import parse_cpp_file, SimTypeFunction, SimTypeBottom
from angr.calling_conventions import DEFAULT_CC
from angr.misc import autoimport
from angr.misc.ux import once
from angr.sim_type import parse_file
from angr.procedures.stubs.ReturnUnconstrained import ReturnUnconstrained
from angr.procedures.stubs.syscall_stub import syscall as stub_syscall

if TYPE_CHECKING:
    from angr.calling_conventions import SimCCSyscall
    from angr.sim_type import SimType


l = logging.getLogger(name=__name__)
SIM_LIBRARIES: dict[str, list[SimLibrary]] = {}
SIM_TYPE_COLLECTIONS: dict[str, SimTypeCollection] = {}


class SimTypeCollection:
    """
    A type collection is the mechanism for describing types. Types in a type collection can be referenced using
    """

    def __init__(self):
        self.names: list[str] | None = None
        self.types: dict[str, SimType] = {}

    def set_names(self, *names: str):
        self.names = list(names)
        for name in names:
            SIM_TYPE_COLLECTIONS[name] = self

    def add(self, name: str, t: SimType) -> None:
        """
        Add a type to the collection.

        :param name:    Name of the type to add.
        :param t:       The SimType object to add to the collection.
        """

        self.types[name] = t

    def get(self, name: str, bottom_on_missing: bool = False) -> SimType:
        """
        Get a SimType object from the collection as identified by the name.

        :param name:    Name of the type to get.
        :param bottom_on_missing:    Return a SimTypeBottom object if the required type does not exist.
        :return:        The SimType object.
        """
        if bottom_on_missing and name not in self.types:
            return SimTypeBottom(label=name)
        if name not in self.types:
            raise AngrMissingTypeError(f"Type {name} is missing")
        return self.types[name]

    def init_str(self) -> str:
        lines = [
            "typelib = SimTypeCollection()",
            "" if not self.names else f"typelib.set_names(*{self.names})",
            "typelib.types = {",
        ]
        for name in sorted(self.types):
            t = self.types[name]
            lines.append(f'    "{name}": {t._init_str()},')
        lines.append("}")

        return "\n".join(lines)

    def __repr__(self):
        return f"<SimTypeCollection with {len(self.types)} types>"


class SimLibrary:
    """
    A SimLibrary is the mechanism for describing a dynamic library's API, its functions and metadata.

    Any instance of this class (or its subclasses) found in the ``angr.procedures.definitions`` package will be
    automatically picked up and added to ``angr.SIM_LIBRARIES`` via all its names.

    :ivar fallback_cc:      A mapping from architecture to the default calling convention that should be used if no
                            other information is present. Contains some sane defaults for linux.
    :ivar fallback_proc:    A SimProcedure class that should be used to provide stub procedures. By default,
                            ``ReturnUnconstrained``.
    """

    def __init__(self):
        self.type_collection_names: list[str] = []
        self.procedures = {}
        self.non_returning = set()
        self.prototypes: dict[str, SimTypeFunction] = {}
        self.default_ccs = {}
        self.names = []
        self.fallback_cc = dict(DEFAULT_CC)
        self.fallback_proc = ReturnUnconstrained

    def copy(self):
        """
        Make a copy of this SimLibrary, allowing it to be mutated without affecting the global version.

        :return:    A new SimLibrary object with the same library references but different dict/list references
        """
        o = SimLibrary()
        o.procedures = dict(self.procedures)
        o.non_returning = set(self.non_returning)
        o.prototypes = dict(self.prototypes)
        o.default_ccs = dict(self.default_ccs)
        o.names = list(self.names)
        return o

    def update(self, other: SimLibrary):
        """
        Augment this SimLibrary with the information from another SimLibrary

        :param other:   The other SimLibrary
        """
        self.procedures.update(other.procedures)
        self.non_returning.update(other.non_returning)
        self.prototypes.update(other.prototypes)
        self.default_ccs.update(other.default_ccs)

    @property
    def name(self):
        """
        The first common name of this library, e.g. libc.so.6, or '??????' if none are known.
        """
        return self.names[0] if self.names else "??????"

    def set_library_names(self, *names):
        """
        Set some common names of this library by which it may be referred during linking

        :param names:   Any number of string library names may be passed as varargs.
        """
        for name in names:
            self.names.append(name)
            if name in SIM_LIBRARIES:
                SIM_LIBRARIES[name].append(self)
            else:
                SIM_LIBRARIES[name] = [self]

    def set_default_cc(self, arch_name, cc_cls):
        """
        Set the default calling convention used for this library under a given architecture

        :param arch_name:   The string name of the architecture, i.e. the ``.name`` field from archinfo.
        :parm cc_cls:       The SimCC class (not an instance!) to use
        """
        arch_name = archinfo.arch_from_id(arch_name).name
        self.default_ccs[arch_name] = cc_cls

    def set_non_returning(self, *names):
        """
        Mark some functions in this class as never returning, i.e. loops forever or terminates execution

        :param names:   Any number of string function names may be passed as varargs
        """
        for name in names:
            self.non_returning.add(name)

    def set_prototype(self, name, proto):
        """
        Set the prototype of a function in the form of a SimTypeFunction containing argument and return types

        :param name:    The name of the function as a string
        :param proto:   The prototype of the function as a SimTypeFunction
        """
        self.prototypes[name] = proto

    def set_prototypes(self, protos):
        """
        Set the prototypes of many functions

        :param protos:   Dictionary mapping function names to SimTypeFunction objects
        """
        self.prototypes.update(protos)

    def set_c_prototype(self, c_decl):
        """
        Set the prototype of a function in the form of a C-style function declaration.

        :param str c_decl: The C-style declaration of the function.
        :return:           A tuple of (function name, function prototype)
        :rtype:            tuple
        """

        parsed = parse_file(c_decl)
        parsed_decl = parsed[0]
        if not parsed_decl:
            raise ValueError("Cannot parse the function prototype.")
        func_name, func_proto = next(iter(parsed_decl.items()))

        self.set_prototype(func_name, func_proto)

        return func_name, func_proto

    def add(self, name, proc_cls, **kwargs):
        """
        Add a function implementation to the library.

        :param name:        The name of the function as a string
        :param proc_cls:    The implementation of the function as a SimProcedure _class_, not instance
        :param kwargs:      Any additional parameters to the procedure class constructor may be passed as kwargs
        """
        self.procedures[name] = proc_cls(display_name=name, **kwargs)

    def add_all_from_dict(self, dictionary, **kwargs):
        """
        Batch-add function implementations to the library.

        :param dictionary:  A mapping from name to procedure class, i.e. the first two arguments to add()
        :param kwargs:      Any additional kwargs will be passed to the constructors of _each_ procedure class
        """
        for name, procedure in dictionary.items():
            self.add(name, procedure, **kwargs)

    def add_alias(self, name, *alt_names):
        """
        Add some duplicate names for a given function. The original function's implementation must already be
        registered.

        :param name:        The name of the function for which an implementation is already present
        :param alt_names:   Any number of alternate names may be passed as varargs
        """
        old_procedure = self.procedures[name]
        for alt in alt_names:
            new_procedure = copy.deepcopy(old_procedure)
            new_procedure.display_name = alt
            self.procedures[alt] = new_procedure
            if name in self.prototypes:
                self.prototypes[alt] = self.prototypes[name]
            if name in self.non_returning:
                self.non_returning.add(alt)

    def _apply_metadata(self, proc, arch):
        if proc.cc is None and arch.name in self.default_ccs:
            proc.cc = self.default_ccs[arch.name](arch)
        if proc.cc is None and arch.name in self.fallback_cc:
            proc.cc = self.fallback_cc[arch.name]["Linux"](arch)
        if proc.display_name in self.prototypes:
            proc.prototype = self.prototypes[proc.display_name].with_arch(arch)
            proc.guessed_prototype = False
            if proc.prototype.arg_names is None:
                # Use inspect to extract the parameters from the run python function
                proc.prototype.arg_names = tuple(inspect.getfullargspec(proc.run).args[1:])
            if not proc.ARGS_MISMATCH:
                proc.num_args = len(proc.prototype.args)
        if proc.display_name in self.non_returning:
            proc.returns = False
        proc.library_name = self.name

    def get(self, name, arch):
        """
        Get an implementation of the given function specialized for the given arch, or a stub procedure if none exists.

        :param name:    The name of the function as a string
        :param arch:    The architecure to use, as either a string or an archinfo.Arch instance
        :return:        A SimProcedure instance representing the function as found in the library
        """
        if type(arch) is str:
            arch = archinfo.arch_from_id(arch)
        if name in self.procedures:
            proc = copy.deepcopy(self.procedures[name])
            self._apply_metadata(proc, arch)
            return proc
        return self.get_stub(name, arch)

    def get_stub(self, name, arch):
        """
        Get a stub procedure for the given function, regardless of if a real implementation is available. This will
        apply any metadata, such as a default calling convention or a function prototype.

        By stub, we pretty much always mean a ``ReturnUnconstrained`` SimProcedure with the appropriate display name
        and metadata set. This will appear in ``state.history.descriptions`` as ``<SimProcedure display_name (stub)>``

        :param name:    The name of the function as a string
        :param arch:    The architecture to use, as either a string or an archinfo.Arch instance
        :return:        A SimProcedure instance representing a plausable stub as could be found in the library.
        """
        proc = self.fallback_proc(display_name=name, is_stub=True)
        self._apply_metadata(proc, arch)
        return proc

    def get_prototype(self, name: str, arch=None) -> SimTypeFunction | None:
        """
        Get a prototype of the given function name, optionally specialize the prototype to a given architecture.

        :param name:    Name of the function.
        :param arch:    The architecture to specialize to.
        :return:        Prototype of the function, or None if the prototype does not exist.
        """
        proto = self.prototypes.get(name, None)
        if proto is None:
            return None
        if arch is not None:
            return proto.with_arch(arch)
        return proto

    def has_metadata(self, name):
        """
        Check if a function has either an implementation or any metadata associated with it

        :param name:    The name of the function as a string
        :return:        A bool indicating if anything is known about the function
        """
        return self.has_implementation(name) or name in self.non_returning or name in self.prototypes

    def has_implementation(self, name):
        """
        Check if a function has an implementation associated with it

        :param name:    The name of the function as a string
        :return:        A bool indicating if an implementation of the function is available
        """
        return name in self.procedures

    def has_prototype(self, func_name):
        """
        Check if a function has a prototype associated with it.

        :param str func_name: The name of the function.
        :return:              A bool indicating if a prototype of the function is available.
        :rtype:               bool
        """

        return func_name in self.prototypes


class SimCppLibrary(SimLibrary):
    """
    SimCppLibrary is a specialized version of SimLibrary that will demangle C++ function names before looking for an
    implementation or prototype for it.
    """

    @staticmethod
    def _try_demangle(name):
        ast = pydemumble.demangle(name)
        return ast if ast else name

    @staticmethod
    def _proto_from_demangled_name(name: str) -> SimTypeFunction | None:
        """
        Attempt to extract arguments and calling convention information for a C++ function whose name was mangled
        according to the Itanium C++ ABI symbol mangling language.

        :param name:    The demangled function name.
        :return:        A prototype or None if a prototype cannot be found.
        """

        try:
            parsed, _ = parse_cpp_file(name, with_param_names=False)
        except ValueError:
            return None
        if not parsed:
            return None
        _, func_proto = next(iter(parsed.items()))
        return func_proto

    def get(self, name, arch):
        """
        Get an implementation of the given function specialized for the given arch, or a stub procedure if none exists.
        Demangle the function name if it is a mangled C++ name.

        :param str name:    The name of the function as a string
        :param arch:    The architecure to use, as either a string or an archinfo.Arch instance
        :return:        A SimProcedure instance representing the function as found in the library
        """
        demangled_name = self._try_demangle(name)
        if demangled_name not in self.procedures:
            return self.get_stub(name, arch)  # get_stub() might use the mangled name to derive the function prototype
        return super().get(demangled_name, arch)

    def get_stub(self, name, arch):
        """
        Get a stub procedure for the given function, regardless of if a real implementation is available. This will
        apply any metadata, such as a default calling convention or a function prototype. Demangle the function name
        if it is a mangled C++ name.

        :param str name:    The name of the function as a string
        :param arch:        The architecture to use, as either a string or an archinfo.Arch instance
        :return:            A SimProcedure instance representing a plausable stub as could be found in the library.
        """
        demangled_name = self._try_demangle(name)
        stub = super().get_stub(demangled_name, arch)
        # try to determine a prototype from the function name if possible
        if demangled_name != name:
            # mangled function name
            stub.prototype = self._proto_from_demangled_name(demangled_name)
            if stub.prototype is not None:
                stub.prototype = stub.prototype.with_arch(arch)
                stub.guessed_prototype = False
                if not stub.ARGS_MISMATCH:
                    stub.num_args = len(stub.prototype.args)
        return stub

    def get_prototype(self, name: str, arch=None) -> SimTypeFunction | None:
        """
        Get a prototype of the given function name, optionally specialize the prototype to a given architecture. The
        function name will be demangled first.

        :param name:    Name of the function.
        :param arch:    The architecture to specialize to.
        :return:        Prototype of the function, or None if the prototype does not exist.
        """
        demangled_name = self._try_demangle(name)
        return super().get_prototype(demangled_name, arch=arch)

    def has_metadata(self, name):
        """
        Check if a function has either an implementation or any metadata associated with it. Demangle the function name
        if it is a mangled C++ name.

        :param name:    The name of the function as a string
        :return:        A bool indicating if anything is known about the function
        """
        name = self._try_demangle(name)
        return super().has_metadata(name)

    def has_implementation(self, name):
        """
        Check if a function has an implementation associated with it. Demangle the function name if it is a mangled C++
        name.

        :param str name:    A mangled function name.
        :return:            bool
        """
        return super().has_implementation(self._try_demangle(name))

    def has_prototype(self, func_name):
        """
        Check if a function has a prototype associated with it. Demangle the function name if it is a mangled C++ name.

        :param str name:    A mangled function name.
        :return:            bool
        """
        return super().has_prototype(self._try_demangle(func_name))


class SimSyscallLibrary(SimLibrary):
    """
    SimSyscallLibrary is a specialized version of SimLibrary for dealing not with a dynamic library's API but rather
    an operating system's syscall API. Because this interface is inherently lower-level than a dynamic library, many
    parts of this class has been changed to store data based on an "ABI name" (ABI = application binary interface,
    like an API but for when there's no programming language) instead of an architecture. An ABI name is just an
    arbitrary string with which a calling convention and a syscall numbering is associated.

    All the SimLibrary methods for adding functions still work, but now there's an additional layer on top that
    associates them with numbers.
    """

    def __init__(self):
        super().__init__()
        self.syscall_number_mapping: dict[str, dict[int, str]] = defaultdict(dict)  # keyed by abi
        self.syscall_name_mapping: dict[str, dict[str, int]] = defaultdict(dict)  # keyed by abi
        self.default_cc_mapping: dict[str, type[SimCCSyscall]] = {}  # keyed by abi
        self.syscall_prototypes: dict[str, dict[str, SimTypeFunction]] = defaultdict(dict)  # keyed by abi
        self.fallback_proc = stub_syscall

    def copy(self):
        o = SimSyscallLibrary()
        o.procedures = dict(self.procedures)
        o.non_returning = set(self.non_returning)
        o.prototypes = dict(self.prototypes)
        o.default_ccs = dict(self.default_ccs)
        o.names = list(self.names)
        o.syscall_number_mapping = defaultdict(dict, self.syscall_number_mapping)  # {abi: {number: name}}
        o.syscall_name_mapping = defaultdict(dict, self.syscall_name_mapping)  # {abi: {name: number}}
        o.syscall_prototypes = defaultdict(dict, self.syscall_prototypes)  # as above
        o.default_cc_mapping = dict(self.default_cc_mapping)  # {abi: cc}
        return o

    def update(self, other):
        super().update(other)
        if isinstance(other, SimSyscallLibrary):
            self.syscall_number_mapping.update(other.syscall_number_mapping)
            self.syscall_name_mapping.update(other.syscall_name_mapping)
            self.default_cc_mapping.update(other.default_cc_mapping)

    def minimum_syscall_number(self, abi):
        """
        :param abi: The abi to evaluate
        :return:    The smallest syscall number known for the given abi
        """
        if abi not in self.syscall_number_mapping or not self.syscall_number_mapping[abi]:
            return 0
        return min(self.syscall_number_mapping[abi])

    def maximum_syscall_number(self, abi):
        """
        :param abi: The abi to evaluate
        :return:    The largest syscall number known for the given abi
        """
        if abi not in self.syscall_number_mapping or not self.syscall_number_mapping[abi]:
            return 0
        return max(self.syscall_number_mapping[abi])

    def add_number_mapping(self, abi, number, name):
        """
        Associate a syscall number with the name of a function present in the underlying SimLibrary

        :param abi:     The abi for which this mapping applies
        :param number:  The syscall number
        :param name:    The name of the function
        """
        self.syscall_number_mapping[abi][number] = name
        self.syscall_name_mapping[abi][name] = number

    def add_number_mapping_from_dict(self, abi, mapping):
        """
        Batch-associate syscall numbers with names of functions present in the underlying SimLibrary

        :param abi:     The abi for which this mapping applies
        :param mapping: A dict mapping syscall numbers to function names
        """
        self.syscall_number_mapping[abi].update(mapping)
        self.syscall_name_mapping[abi].update({b: a for a, b in mapping.items()})

    def set_abi_cc(self, abi, cc_cls):
        """
        Set the default calling convention for an abi

        :param abi:     The name of the abi
        :param cc_cls:  A SimCC _class_, not an instance, that should be used for syscalls using the abi
        """
        self.default_cc_mapping[abi] = cc_cls

    def set_prototype(self, abi: str, name: str, proto: SimTypeFunction) -> None:  # pylint: disable=arguments-differ
        """
        Set the prototype of a function in the form of a SimTypeFunction containing argument and return types

        :param abi:     ABI of the syscall.
        :param name:    The name of the syscall as a string
        :param proto:   The prototype of the syscall as a SimTypeFunction
        """
        self.syscall_prototypes[abi][name] = proto

    def set_prototypes(self, abi: str, protos: dict[str, SimTypeFunction]) -> None:  # pylint: disable=arguments-differ
        """
        Set the prototypes of many syscalls.

        :param abi:     ABI of the syscalls.
        :param protos:  Dictionary mapping syscall names to SimTypeFunction objects
        """
        self.syscall_prototypes[abi].update(protos)

    def _canonicalize(self, number, arch, abi_list):
        if type(arch) is str:
            arch = archinfo.arch_from_id(arch)
        if type(number) is str:
            return number, arch, None
        for abi in abi_list:
            mapping = self.syscall_number_mapping[abi]
            if number in mapping:
                return mapping[number], arch, abi
        return f"sys_{number}", arch, None

    def _apply_numerical_metadata(self, proc, number, arch, abi):
        proc.syscall_number = number
        proc.abi = abi
        if abi in self.default_cc_mapping:
            cc = self.default_cc_mapping[abi](arch)
            proc.cc = cc
        # a bit of a hack.
        name = proc.display_name
        if self.syscall_prototypes[abi].get(name, None) is not None:
            proc.guessed_prototype = False
            proc.prototype = self.syscall_prototypes[abi][name].with_arch(arch)

    # pylint: disable=arguments-differ
    def get(self, number, arch, abi_list=()):
        """
        The get() function for SimSyscallLibrary looks a little different from its original version.

        Instead of providing a name, you provide a number, and you additionally provide a list of abi names that are
        applicable. The first abi for which the number is present in the mapping will be chosen. This allows for the
        easy abstractions of architectures like ARM or MIPS linux for which there are many ABIs that can be used at any
        time by using syscall numbers from various ranges. If no abi knows about the number, the stub procedure with
        the name "sys_%d" will be used.

        :param number:      The syscall number
        :param arch:        The architecture being worked with, as either a string name or an archinfo.Arch
        :param abi_list:    A list of ABI names that could be used
        :return:            A SimProcedure representing the implementation of the given syscall, or a stub if no
                            implementation is available
        """
        name, arch, abi = self._canonicalize(number, arch, abi_list)
        proc = super().get(name, arch)
        proc.is_syscall = True
        self._apply_numerical_metadata(proc, number, arch, abi)
        return proc

    def get_stub(self, number, arch, abi_list=()):
        """
        Pretty much the intersection of SimLibrary.get_stub() and SimSyscallLibrary.get().

        :param number:      The syscall number
        :param arch:        The architecture being worked with, as either a string name or an archinfo.Arch
        :param abi_list:    A list of ABI names that could be used
        :return:            A SimProcedure representing a plausable stub that could model the syscall
        """
        name, arch, abi = self._canonicalize(number, arch, abi_list)
        proc = super().get_stub(name, arch)
        self._apply_numerical_metadata(proc, number, arch, abi)
        l.debug("unsupported syscall: %s", number)
        return proc

    def get_prototype(self, abi: str, name: str, arch=None) -> SimTypeFunction | None:
        """
        Get a prototype of the given syscall name and its ABI, optionally specialize the prototype to a given
        architecture.

        :param abi:     ABI of the prototype to get.
        :param name:    Name of the syscall.
        :param arch:    The architecture to specialize to.
        :return:        Prototype of the syscall, or None if the prototype does not exist.
        """
        if abi not in self.syscall_prototypes:
            return None
        proto = self.syscall_prototypes[abi].get(name, None)
        if proto is None:
            return None
        return proto.with_arch(arch=arch)

    def has_metadata(self, number, arch, abi_list=()):
        """
        Pretty much the intersection of SimLibrary.has_metadata() and SimSyscallLibrary.get().

        :param number:      The syscall number
        :param arch:        The architecture being worked with, as either a string name or an archinfo.Arch
        :param abi_list:    A list of ABI names that could be used
        :return:            A bool of whether or not any implementation or metadata is known about the given syscall
        """
        name, _, _ = self._canonicalize(number, arch, abi_list)
        return super().has_metadata(name)

    def has_implementation(self, number, arch, abi_list=()):
        """
        Pretty much the intersection of SimLibrary.has_implementation() and SimSyscallLibrary.get().

        :param number:      The syscall number
        :param arch:        The architecture being worked with, as either a string name or an archinfo.Arch
        :param abi_list:    A list of ABI names that could be used
        :return:            A bool of whether or not an implementation of the syscall is available
        """
        name, _, _ = self._canonicalize(number, arch, abi_list)
        return super().has_implementation(name)

    def has_prototype(self, abi: str, name: str) -> bool:
        """
        Check if a function has a prototype associated with it. Demangle the function name if it is a mangled C++ name.

        :param abi:         Name of the ABI.
        :param name:        The syscall name.
        :return:            bool
        """
        if abi not in self.syscall_prototypes:
            return False
        return name in self.syscall_prototypes[abi]


#
# Autoloading
#

# By default we only load common API definitions (as defined in COMMON_LIBRARIES). For loading more definitions, the
# following logic is followed:
# - We will load all Windows APIs them if the loaded binary is a Windows binary, or when load_win32api_definitions() is
#   called.
# - We will load all APIs when load_all_definitions() is called.

_DEFINITIONS_BASEDIR = os.path.dirname(os.path.realpath(__file__))
_EXTERNAL_DEFINITIONS_DIRS: list[str] | None = None


def load_type_collections(skip=None) -> None:
    if skip is None:
        skip = set()

    for _ in autoimport.auto_import_modules(
        "angr.procedures.definitions",
        _DEFINITIONS_BASEDIR,
        filter_func=lambda module_name: module_name.startswith("types_") and module_name not in skip,
    ):
        pass


def load_win32_type_collections() -> None:
    if once("load_win32_type_collections"):
        for _ in autoimport.auto_import_modules(
            "angr.procedures.definitions",
            _DEFINITIONS_BASEDIR,
            filter_func=lambda module_name: module_name == "types_win32",
        ):
            pass


def load_external_definitions():
    """
    Load library definitions from specific directories. By default it parses ANGR_EXTERNAL_DEFINITIONS_DIRS as a
    semicolon separated list of directory paths. Then it loads all .py files in each directory. These .py files should
    declare SimLibrary() objects and call .set_library_names() to register themselves in angr.SIM_LIBRARIES.
    """

    global _EXTERNAL_DEFINITIONS_DIRS

    if _EXTERNAL_DEFINITIONS_DIRS is None and "ANGR_EXTERNAL_DEFINITIONS_DIRS" in os.environ:
        _EXTERNAL_DEFINITIONS_DIRS = os.environ["ANGR_EXTERNAL_DEFINITIONS_DIRS"].strip('"').split(";")
        l.debug("Using external library definitions from %s", _EXTERNAL_DEFINITIONS_DIRS)
        for d in _EXTERNAL_DEFINITIONS_DIRS:
            if not os.path.isdir(d):
                l.warning("External library definitions directory %s does not exist or is not a directory.", d)

    if _EXTERNAL_DEFINITIONS_DIRS:
        # we must load all definitions prior to any external definitions are loaded. otherwise external definitions may
        # be overwritten by embedded definitions in angr, which is undesirable
        load_all_definitions()

        for d in _EXTERNAL_DEFINITIONS_DIRS:
            for _ in autoimport.auto_import_source_files(d):
                pass


def load_win32api_definitions():
    load_win32_type_collections()
    if once("load_win32api_definitions"):
        for _ in autoimport.auto_import_modules(
            "angr.procedures.definitions",
            _DEFINITIONS_BASEDIR,
            filter_func=lambda module_name: module_name.startswith(("win32_", "wdk_"))
            or module_name in {"ntoskrnl", "ntdll", "user32"},
        ):
            pass


def load_all_definitions():
    load_type_collections(skip=set())
    if once("load_all_definitions"):
        for _ in autoimport.auto_import_modules("angr.procedures.definitions", _DEFINITIONS_BASEDIR):
            pass


COMMON_LIBRARIES = {
    # CGC
    "cgc",
    # (mostly) Linux
    "glibc",
    "gnulib",  # really just for .o files in coreutils
    "libstdcpp",
    "linux_kernel",
    "linux_loader",
    # Windows
    "msvcr",
}


# Load common types
load_type_collections(skip={"types_win32"})


# Load common definitions
for _ in autoimport.auto_import_modules(
    "angr.procedures.definitions",
    _DEFINITIONS_BASEDIR,
    filter_func=lambda module_name: module_name in COMMON_LIBRARIES,
):
    pass



# pylint:disable=unused-argument
from __future__ import annotations
from typing import TYPE_CHECKING
import logging

from angr.ailment import Block, AILBlockWalker
from angr.ailment.expression import Load, Const, BinaryOp, UnaryOp
from angr.ailment.statement import Statement, Assignment, Call, ConditionalJump

from angr.analyses.decompiler.ailgraph_walker import AILGraphWalker
from .optimization_pass import OptimizationPass, OptimizationPassStage

if TYPE_CHECKING:
    from angr import Project


_l = logging.getLogger(name=__name__)


class BlockWalker(AILBlockWalker):
    def __init__(self, project: Project):
        super().__init__()
        self._project = project
        self._new_block: Block | None = None  # output

    def walk(self, block: Block):
        self._new_block = None
        super().walk(block)
        return self._new_block

    def _addr_belongs_to_ro_region(self, addr: int) -> bool:
        section = self._project.loader.find_section_containing(addr)
        if section is not None:
            return not section.is_writable
        segment = self._project.loader.find_segment_containing(addr)
        if segment is not None:
            return not segment.is_writable
        return False

    def _addr_belongs_to_got(self, addr: int) -> bool:
        section = self._project.loader.find_section_containing(addr)
        if section is not None:
            return section.name and "got" in section.name
        return False

    def _addr_belongs_to_object(self, addr: int) -> bool:
        obj = self._project.loader.find_object_containing(addr)
        return obj is not None

    def _handle_stmt(self, stmt_idx: int, stmt: Statement, block: Block):
        r = super()._handle_stmt(stmt_idx, stmt, block)
        if r is not None:
            # replace the original statement
            if self._new_block is None:
                self._new_block = block.copy()
            self._new_block.statements[stmt_idx] = r

    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block):
        new_dst = self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
        new_src = self._handle_expr(1, stmt.src, stmt_idx, stmt, block)

        if new_dst is not None or new_src is not None:
            return Assignment(
                stmt.idx,
                stmt.dst if new_dst is None else new_dst,
                stmt.src if new_src is None else new_src,
                **stmt.tags,
            )
        return None

    def _handle_CallExpr(self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement, block: Block | None):
        new_target = self._handle_expr(-1, expr.target, stmt_idx, stmt, block)

        new_args = None
        if expr.args:
            i = 0
            new_exprs = []
            while i < len(expr.args):
                arg = expr.args[i]
                new_expr = self._handle_expr(i, arg, stmt_idx, stmt, block)
                new_exprs.append(new_expr)
                i += 1
            if any(expr_ is not None for expr_ in new_exprs):
                # create new args
                new_args = [
                    (new_arg if new_arg is not None else old_arg) for new_arg, old_arg in zip(new_exprs, expr.args)
                ]

        if new_target is not None or new_args is not None:
            # create a new call expr
            return Call(
                expr.idx,
                expr.target if new_target is None else new_target,
                calling_convention=expr.calling_convention,
                prototype=expr.prototype,
                args=new_args if new_args is not None else expr.args,
                ret_expr=expr.ret_expr,
                **expr.tags,
            )
        return None

    def _handle_Call(self, stmt_idx: int, stmt: Call, block: Block):
        new_target = self._handle_expr(-1, stmt.target, stmt_idx, stmt, block)

        new_args = None
        if stmt.args:
            i = 0
            new_exprs = []
            while i < len(stmt.args):
                arg = stmt.args[i]
                new_expr = self._handle_expr(i, arg, stmt_idx, stmt, block)
                new_exprs.append(new_expr)
                i += 1
            if any(expr is not None for expr in new_exprs):
                # create new args
                new_args = [
                    (new_arg if new_arg is not None else old_arg) for new_arg, old_arg in zip(new_exprs, stmt.args)
                ]

        if new_target is not None or new_args is not None:
            # create a new statement
            return Call(
                stmt.idx,
                stmt.target if new_target is None else new_target,
                calling_convention=stmt.calling_convention,
                prototype=stmt.prototype,
                args=new_args if new_args is not None else stmt.args,
                ret_expr=stmt.ret_expr,
                **stmt.tags,
            )
        return None

    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement, block: Block):
        if isinstance(expr.addr, Const):
            # *(const_addr)
            # does it belong to a read-only section/segment?
            is_got = self._addr_belongs_to_got(expr.addr.value)
            if is_got or self._addr_belongs_to_ro_region(expr.addr.value):
                try:
                    w = self._project.loader.memory.unpack_word(
                        expr.addr.value,
                        expr.addr.bits // self._project.arch.byte_width,
                        endness=self._project.arch.memory_endness,
                    )
                except KeyError:
                    # we don't have enough bytes to read out
                    w = None
                if w is not None and not (is_got and w == 0):
                    # nice! replace it with the actual value
                    return Const(None, None, w, expr.bits, **expr.tags)
        elif (
            isinstance(expr.addr, Load)
            and expr.addr.bits == self._project.arch.bits
            and isinstance(expr.addr.addr, Const)
            and (
                # *(*(const_addr))
                # does it belong to a read-only section/segment?
                self._addr_belongs_to_got(expr.addr.addr.value)
                or self._addr_belongs_to_ro_region(expr.addr.addr.value)
            )
        ):
            w = self._project.loader.memory.unpack_word(
                expr.addr.addr.value,
                expr.addr.addr.bits // self._project.arch.byte_width,
                endness=self._project.arch.memory_endness,
            )
            if w is not None and self._addr_belongs_to_object(w):
                # nice! replace it with a load from that address
                return Load(
                    expr.idx,
                    Const(None, None, w, expr.addr.size, **expr.addr.addr.tags),
                    expr.size,
                    expr.endness,
                    variable=expr.variable,
                    variable_offset=expr.variable_offset,
                    guard=expr.guard,
                    alt=expr.alt,
                    **expr.tags,
                )

        return super()._handle_Load(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Block):
        new_operands = [
            self._handle_expr(0, expr.operands[0], stmt_idx, stmt, block),
            self._handle_expr(1, expr.operands[1], stmt_idx, stmt, block),
        ]
        if any(op is not None for op in new_operands):
            new_operands = [
                (new_op if new_op is not None else old_op) for new_op, old_op in zip(new_operands, expr.operands)
            ]
            return BinaryOp(
                expr.idx,
                expr.op,
                new_operands,
                expr.signed,
                variable=expr.variable,
                variable_offset=expr.variable_offset,
                floating_point=expr.floating_point,
                rounding_mode=expr.rounding_mode,
                **expr.tags,
            )
        return None

    def _handle_UnaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Block):
        new_operand = self._handle_expr(0, expr.operands[0], stmt_idx, stmt, block)
        if new_operand is not None:
            return UnaryOp(
                expr.idx,
                expr.op,
                new_operand,
                expr.signed,
                variable=expr.variable,
                variable_offset=expr.variable_offset,
                bits=expr.bits,
                **expr.tags,
            )
        return None

    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Block):
        new_cond = self._handle_expr(0, stmt.condition, stmt_idx, stmt, block)
        new_true_target = self._handle_expr(1, stmt.true_target, stmt_idx, stmt, block)
        new_false_target = self._handle_expr(2, stmt.false_target, stmt_idx, stmt, block)

        if new_cond is not None or new_true_target is not None or new_false_target is not None:
            return ConditionalJump(
                stmt.idx,
                new_cond if new_cond is not None else stmt.condition,
                new_true_target if new_true_target is not None else stmt.true_target,
                new_false_target if new_false_target is not None else stmt.false_target,
                true_target_idx=stmt.true_target_idx,
                false_target_idx=stmt.false_target_idx,
                **stmt.tags,
            )
        return None


class ConstantDereferencesSimplifier(OptimizationPass):
    """
    Makes the following simplifications::

        *(*(const_addr))  ==>  *(value) iff  *const_addr == value
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.AFTER_SINGLE_BLOCK_SIMPLIFICATION
    NAME = "Simplify constant dereferences"
    DESCRIPTION = __doc__.strip()

    def __init__(self, func, **kwargs):
        super().__init__(func, **kwargs)

        self._block_walker = BlockWalker(self.project)
        self.analyze()

    def _check(self):
        return True, None

    def _analyze(self, cache=None):
        # walk the entire graph and traverse each expression
        walker = AILGraphWalker(self._graph, handler=self._walk_block, replace_nodes=True)
        walker.walk()

    def _walk_block(self, block: Block) -> Block | None:
        return self._block_walker.walk(block)



from __future__ import annotations
import angr


class NoReturnUnconstrained(angr.SimProcedure):  # pylint:disable=redefined-builtin
    """
    Use in places you would put ReturnUnconstrained as a default action but the function shouldn't actually return.
    """

    NO_RET = True

    def run(self, **kwargs):  # pylint:disable=unused-argument
        self.exit(self.state.solver.Unconstrained("unconstrained_exit_code", self.state.arch.bits))



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("api-ms-win-core-file-fromapp-l1-1-0.dll")
prototypes = \
    {
        #
        'CopyFileFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpExistingFileName", "lpNewFileName", "bFailIfExists"]),
        #
        'CreateDirectoryFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("SECURITY_ATTRIBUTES", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpPathName", "lpSecurityAttributes"]),
        #
        'CreateFileFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("SECURITY_ATTRIBUTES", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["lpFileName", "dwDesiredAccess", "dwShareMode", "lpSecurityAttributes", "dwCreationDisposition", "dwFlagsAndAttributes", "hTemplateFile"]),
        #
        'CreateFile2FromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("CREATEFILE2_EXTENDED_PARAMETERS", SimStruct), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["lpFileName", "dwDesiredAccess", "dwShareMode", "dwCreationDisposition", "pCreateExParams"]),
        #
        'DeleteFileFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpFileName"]),
        #
        'FindFirstFileExFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="FINDEX_INFO_LEVELS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="FINDEX_SEARCH_OPS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["lpFileName", "fInfoLevelId", "lpFindFileData", "fSearchOp", "lpSearchFilter", "dwAdditionalFlags"]),
        #
        'GetFileAttributesExFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="GET_FILEEX_INFO_LEVELS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpFileName", "fInfoLevelId", "lpFileInformation"]),
        #
        'MoveFileFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpExistingFileName", "lpNewFileName"]),
        #
        'RemoveDirectoryFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpPathName"]),
        #
        'ReplaceFileFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpReplacedFileName", "lpReplacementFileName", "lpBackupFileName", "dwReplaceFlags", "lpExclude", "lpReserved"]),
        #
        'SetFileAttributesFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpFileName", "dwFileAttributes"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations
import struct
import zlib
from io import BytesIO

from angr.errors import AngrError
from .consts import FlirtParseFlag, FlirtFeatureFlag, FlirtFunctionFlag
from .flirt_utils import read_max_2_bytes, read_multiple_bytes
from .flirt_function import FlirtFunction
from .flirt_module import FlirtModule
from .flirt_node import FlirtNode


class FlirtSignatureError(AngrError):
    """
    Describes errors related to FLIRT signatures, especially parsing.
    """


class FlirtSignature:
    """
    This class describes a FLIRT signature without any internal data that is only available after parsing.
    """

    def __init__(
        self,
        arch: str,
        platform: str,
        sig_name: str,
        sig_path: str,
        unique_strings: set[str] | None = None,
        compiler: str | None = None,
        compiler_version: str | None = None,
        os_name: str | None = None,
        os_version: str | None = None,
    ):
        self.arch = arch
        self.platform = platform
        self.sig_name = sig_name
        self.sig_path = sig_path
        self.unique_strings = unique_strings
        self.compiler = compiler
        self.compiler_version = compiler_version
        self.os_name = os_name
        self.os_version = os_version

    def __repr__(self):
        if self.os_name:
            if self.os_version:
                return f"<{self.sig_name}@{self.arch}-{self.os_name}-{self.os_version}>"
            return f"<{self.sig_name}@{self.arch}-{self.os_name}>"
        return f"<{self.sig_name}@{self.arch}-{self.platform}>"


class FlirtSignatureParsed:
    """
    Describes a FLIRT signature file after parsing.
    """

    __slots__ = (
        "app_types",
        "arch",
        "crc",
        "ctype",
        "ctypes_crc",
        "features",
        "file_types",
        "libname",
        "nfuncs",
        "os_types",
        "pattern_size",
        "root",
        "version",
    )

    def __init__(
        self,
        version: int,
        arch: int,
        file_types: int,
        os_types: int,
        app_types: int,
        features: int,
        crc: int,
        ctype: int,
        ctypes_crc: int,
        nfuncs: int | None,
        pattern_size: int | None,
        libname: str,
        root: FlirtNode | None,
    ):
        self.version = version
        self.arch = arch
        self.file_types = file_types
        self.os_types = os_types
        self.app_types = app_types
        self.features = features
        self.crc = crc
        self.ctype = ctype
        self.ctypes_crc = ctypes_crc
        self.nfuncs = nfuncs
        self.pattern_size = pattern_size
        self.libname = libname
        self.root = root

    def parse_tree(self, file_obj, root: bool = False) -> FlirtNode:
        """
        Parse a FLIRT function tree.
        """

        if not root:
            length = file_obj.read(1)[0]
            variant_mask = self.parse_variant_mask(file_obj, length)
            pattern = self.parse_node(file_obj, length, variant_mask)
        else:
            length = 0
            pattern = []

        node_count = read_multiple_bytes(file_obj)
        if node_count > 0:
            # non-leaf; load its child nodes
            nodes: list[FlirtNode] = [None] * node_count  # type: ignore
            for i in range(node_count):
                nn = self.parse_tree(file_obj)
                nodes[i] = nn
            return FlirtNode(nodes, [], length, pattern)
        # leaf
        modules = self.parse_modules(file_obj)
        return FlirtNode([], modules, length, pattern)

    def parse_public_function(self, file_obj, offset: int) -> tuple[FlirtFunction, int, int]:
        off = read_multiple_bytes(file_obj) if self.version >= 9 else read_max_2_bytes(file_obj)
        off += offset

        local = False  # is it a local function?
        collision = False  # is it an unresolved collision?

        flags = file_obj.read(1)[0]
        if flags < 0x20:
            local = bool(flags & FlirtFunctionFlag.FUNCTION_LOCAL)
            collision = bool(flags & FlirtFunctionFlag.FUNCTION_UNRESOLVED_COLLISION)
            next_byte = file_obj.read(1)[0]
        else:
            next_byte = flags

        name_lst = []
        name_end = False  # in case the function name is too long...
        for _ in range(1024):  # max length of a function name
            if next_byte < 0x20 or next_byte >= 0x80:
                name_end = True
                break
            name_lst.append(next_byte)
            next_byte = file_obj.read(1)[0]

        name = bytes(name_lst).decode("utf-8")
        if not name_end:
            name = name + "..."
        return FlirtFunction(name, off, local, collision), off, next_byte

    def parse_referenced_functions(self, file_obj) -> list[FlirtFunction]:
        func_count = file_obj.read(1)[0] if self.version >= 8 else 1
        lst = []
        for _ in range(func_count):
            off = read_multiple_bytes(file_obj) if self.version >= 9 else read_max_2_bytes(file_obj)
            name_len = file_obj.read(1)[0]
            if name_len == 0:
                name_len = read_multiple_bytes(file_obj)
            if name_len > 1024:
                raise FlirtSignatureError(f"Function name too long: {name_len}")
            name_bytes = file_obj.read(name_len)
            if len(name_bytes) < name_len:
                raise FlirtSignatureError("Unexpected EOF")
            name = name_bytes.decode("utf-8").rstrip("\x00")
            lst.append(FlirtFunction(name, off, False, False))
        return lst

    def parse_tail_bytes(self, file_obj) -> list[tuple[int, int]]:
        bytes_count = file_obj.read(1)[0] if self.version >= 8 else 1
        lst = []
        for _ in range(bytes_count):
            off = read_multiple_bytes(file_obj) if self.version >= 9 else read_max_2_bytes(file_obj)
            value = file_obj.read(1)[0]
            lst.append((off, value))
        return lst

    def parse_modules(self, file_obj) -> list[FlirtModule]:
        modules = []
        while True:
            crc_len = file_obj.read(1)[0]
            crc = struct.unpack(">H", file_obj.read(2))[0]

            while True:
                # parse all modules with the same CRC
                module, flags = self.parse_module(file_obj)
                module.crc_len = crc_len
                module.crc = crc
                modules.append(module)
                if flags & FlirtParseFlag.PARSE_MORE_MODULES_WITH_SAME_CRC == 0:
                    break

            # same crc length but different crc
            if flags & FlirtParseFlag.PARSE_MORE_MODULES == 0:
                break
        return modules

    def parse_module(self, file_obj) -> tuple[FlirtModule, int]:
        length = read_multiple_bytes(file_obj) if self.version >= 9 else read_max_2_bytes(file_obj)
        pub_funcs = []
        off = 0
        while True:
            func, off, flags = self.parse_public_function(file_obj, off)
            pub_funcs.append(func)
            if flags & FlirtParseFlag.PARSE_MORE_PUBLIC_NAMES == 0:
                break

        tail_bytes: list[tuple[int, int]] = []
        if flags & FlirtParseFlag.PARSE_READ_TAIL_BYTES:
            tail_bytes = self.parse_tail_bytes(file_obj)

        ref_funcs = []
        if flags & FlirtParseFlag.PARSE_READ_REFERENCED_FUNCTIONS:
            ref_funcs = self.parse_referenced_functions(file_obj)

        return (
            FlirtModule(
                length,
                0,  # back-filled in its caller
                0,  # back-filled in its caller
                pub_funcs,
                ref_funcs,
                tail_bytes,
            ),
            flags,
        )

    @staticmethod
    def parse_variant_mask(file_obj, length: int) -> int:
        if length < 0x10:
            return read_max_2_bytes(file_obj)
        if length <= 0x20:
            return read_multiple_bytes(file_obj)
        if length <= 0x40:
            return (read_multiple_bytes(file_obj) << 32) | read_multiple_bytes(file_obj)
        raise FlirtSignatureError(f"Unexpected variant mask length: {length}")

    @staticmethod
    def is_bit_set_be(mask: int, mask_len: int, bit_offset: int) -> bool:
        assert mask_len > bit_offset
        return mask & (1 << (mask_len - bit_offset - 1)) != 0

    @staticmethod
    def parse_node(file_obj, length: int, variant_mask: int) -> list[int]:
        pattern = [-1] * length
        for i in range(length):
            if not FlirtSignatureParsed.is_bit_set_be(variant_mask, length, i):
                pattern[i] = file_obj.read(1)[0]
        return pattern

    @classmethod
    def parse(cls, file_obj) -> FlirtSignatureParsed:
        """
        Parse a FLIRT signature file.

        The following struct definitions come from radare2

        // FLIRT v5+
        ut8 magic[6];
        ut8 version;
        ut8 arch;
        ut32 file_types;
        ut16 os_types;
        ut16 app_types;
        ut16 features;
        ut16 old_n_functions;
        ut16 crc16;
        ut8 ctype[12];
        ut8 library_name_len;
        ut16 ctypes_crc16;

        // FLIRT v6+
        ut32 nfuncs;

        // FLIRT v8+
        ut16 pattern_size;

        // FLIRT v10
        ut16 unknown;
        """

        struct_str = "<6s B B I H H H H H 12s B H".replace(" ", "")
        sz = struct.calcsize(struct_str)
        header_bytes = file_obj.read(sz)
        if len(header_bytes) != sz:
            raise FlirtSignatureError
        unpacked = struct.unpack(struct_str, header_bytes)

        # sanity check
        if unpacked[0] != b"IDASGN":
            raise FlirtSignatureError("Unexpected magic bytes")

        version = unpacked[1]
        if version < 5:
            raise FlirtSignatureError("Unsupported FLIRT signature version")

        if version >= 6:
            # nfuncs
            data = file_obj.read(4)
            if len(data) != 4:
                raise FlirtSignatureError("Unexpected EOF")
            nfuncs = struct.unpack("<I", data)[0]
        else:
            nfuncs = None
        if version >= 8:
            # pattern_size
            data = file_obj.read(2)
            if len(data) != 2:
                raise FlirtSignatureError("Unexpected EOF")
            pattern_size = struct.unpack("<H", data)[0]
        else:
            pattern_size = None

        if version >= 10:
            # unknown
            data = file_obj.read(2)
            if len(data) != 2:
                raise FlirtSignatureError("Unexpected EOF")

        libname_len = unpacked[10]
        libname = file_obj.read(libname_len).decode("utf-8")

        obj = cls(
            version=version,
            arch=unpacked[2],
            file_types=unpacked[3],
            os_types=unpacked[4],
            app_types=unpacked[5],
            features=unpacked[6],
            crc=unpacked[7],
            ctype=unpacked[8],
            ctypes_crc=unpacked[11],
            nfuncs=nfuncs,
            pattern_size=pattern_size,
            libname=libname,
            root=None,
        )

        # is it compressed?
        if obj.features & FlirtFeatureFlag.FEATURE_COMPRESSED:
            data = file_obj.read()
            try:
                decompressed = BytesIO(zlib.decompress(data))
            except zlib.error as ex:
                raise FlirtSignatureError(f"Failed to decompress FLIRT signature: {ex}") from ex
            file_obj = decompressed

        root = obj.parse_tree(file_obj, root=True)

        obj.root = root
        return obj



from __future__ import annotations
import claripy

import angr


class fork(angr.SimProcedure):
    def run(self):
        return claripy.If(
            claripy.BoolS("fork_parent"),
            claripy.BVV(1338, self.arch.sizeof["int"]),
            claripy.BVV(0, self.arch.sizeof["int"]),
        )



from __future__ import annotations
from collections import OrderedDict

from angr.ailment.statement import Call, Store, ConditionalJump, CAS
from angr.ailment.expression import (
    Register,
    BinaryOp,
    StackBaseOffset,
    ITE,
    VEXCCallExpression,
    Tmp,
    DirtyExpression,
    Load,
)

from angr.engines.light import SimEngineLightAIL
from angr.project import Project
from angr.utils.ssa import get_reg_offset_base
from angr.utils.orderedset import OrderedSet
from angr.calling_conventions import default_cc
from .traversal_state import TraversalState


class SimEngineSSATraversal(SimEngineLightAIL[TraversalState, None, None, None]):
    """
    This engine collects all register and stack variable locations and links them to the block of their creation.
    """

    def __init__(
        self,
        project: Project,
        simos,
        sp_tracker=None,
        bp_as_gpr: bool = False,
        def_to_loc=None,
        loc_to_defs=None,
        stackvars: bool = False,
        use_tmps: bool = False,
    ):
        super().__init__(project)
        self.simos = simos
        self.sp_tracker = sp_tracker
        self.bp_as_gpr = bp_as_gpr
        self.stackvars = stackvars
        self.use_tmps = use_tmps

        self.def_to_loc = def_to_loc if def_to_loc is not None else []
        self.loc_to_defs = loc_to_defs if loc_to_defs is not None else OrderedDict()

    def _is_top(self, expr):
        return True

    def _top(self, bits):
        return None

    def _process_block_end(self, block, stmt_data, whitelist):
        pass

    def _handle_stmt_Assignment(self, stmt):
        if isinstance(stmt.dst, Register):
            codeloc = self._codeloc()
            self.def_to_loc.append((stmt.dst, codeloc))
            if codeloc not in self.loc_to_defs:
                self.loc_to_defs[codeloc] = OrderedSet()
            self.loc_to_defs[codeloc].add(stmt.dst)

            base_off = get_reg_offset_base(stmt.dst.reg_offset, self.arch)
            self.state.live_registers.add(base_off)

        self._expr(stmt.src)

    def _handle_stmt_WeakAssignment(self, stmt):
        self._expr(stmt.src)
        self._expr(stmt.dst)

    def _handle_stmt_CAS(self, stmt: CAS):
        self._expr(stmt.addr)
        self._expr(stmt.data_lo)
        if stmt.data_hi is not None:
            self._expr(stmt.data_hi)
        self._expr(stmt.expd_lo)
        if stmt.expd_hi is not None:
            self._expr(stmt.expd_hi)

    def _handle_stmt_Store(self, stmt: Store):
        self._expr(stmt.addr)
        self._expr(stmt.data)
        if stmt.guard is not None:
            self._expr(stmt.guard)

        if self.stackvars and isinstance(stmt.addr, StackBaseOffset) and isinstance(stmt.addr.offset, int):
            codeloc = self._codeloc()
            self.def_to_loc.append((stmt, codeloc))
            if codeloc not in self.loc_to_defs:
                self.loc_to_defs[codeloc] = OrderedSet()
            self.loc_to_defs[codeloc].add(stmt)

            self.state.live_stackvars.add((stmt.addr.offset, stmt.size))

    def _handle_stmt_ConditionalJump(self, stmt: ConditionalJump):
        self._expr(stmt.condition)
        if stmt.true_target is not None:
            self._expr(stmt.true_target)
        if stmt.false_target is not None:
            self._expr(stmt.false_target)

    def _handle_stmt_Call(self, stmt: Call):

        # kill caller-saved registers
        cc = (
            default_cc(self.arch.name, platform=self.simos.name if self.simos is not None else None)
            if stmt.calling_convention is None
            else stmt.calling_convention
        )
        assert cc is not None
        for reg_name in cc.CALLER_SAVED_REGS:
            reg_offset = self.arch.registers[reg_name][0]
            base_off = get_reg_offset_base(reg_offset, self.arch)
            self.state.live_registers.discard(base_off)

        if stmt.ret_expr is not None and isinstance(stmt.ret_expr, Register):
            codeloc = self._codeloc()
            self.def_to_loc.append((stmt.ret_expr, codeloc))
            if codeloc not in self.loc_to_defs:
                self.loc_to_defs[codeloc] = OrderedSet()
            self.loc_to_defs[codeloc].add(stmt.ret_expr)

            base_off = get_reg_offset_base(stmt.ret_expr.reg_offset, self.arch)
            self.state.live_registers.add(base_off)

    def _handle_stmt_Dummy(self, stmt):
        pass

    def _handle_stmt_DirtyStatement(self, stmt):
        self._expr(stmt.dirty)

    def _handle_stmt_Jump(self, stmt):
        self._expr(stmt.target)

    _handle_stmt_Label = _handle_stmt_Dummy

    def _handle_stmt_Return(self, stmt):
        for expr in stmt.ret_exprs:
            self._expr(expr)

    def _handle_expr_Register(self, expr: Register):
        base_offset = get_reg_offset_base(expr.reg_offset, self.arch)

        if base_offset not in self.state.live_registers:
            codeloc = self._codeloc()
            self.def_to_loc.append((expr, codeloc))
            if codeloc not in self.loc_to_defs:
                self.loc_to_defs[codeloc] = OrderedSet()
            self.loc_to_defs[codeloc].add(expr)

            self.state.live_registers.add(base_offset)

    def _handle_expr_Load(self, expr: Load):
        self._expr(expr.addr)
        if (
            self.stackvars
            and isinstance(expr.addr, StackBaseOffset)
            and isinstance(expr.addr.offset, int)
            and (expr.addr.offset, expr.size) not in self.state.live_stackvars
        ):
            # we must create this stack variable on the fly; we did not see its creation before it is first used
            codeloc = self._codeloc()
            self.def_to_loc.append((expr, codeloc))
            if codeloc not in self.loc_to_defs:
                self.loc_to_defs[codeloc] = OrderedSet()
            self.loc_to_defs[codeloc].add(expr)
            self.state.live_stackvars.add((expr.addr.offset, expr.size))

    def _handle_expr_StackBaseOffset(self, expr: StackBaseOffset):
        # we don't know the size, so we assume the size is 1 for now...
        sz = 1
        if isinstance(expr.offset, int) and (expr.offset, sz) not in self.state.live_stackvars:
            codeloc = self._codeloc()
            self.def_to_loc.append((expr, codeloc))
            if codeloc not in self.loc_to_defs:
                self.loc_to_defs[codeloc] = OrderedSet()
            self.loc_to_defs[codeloc].add(expr)
            self.state.live_stackvars.add((expr.offset, sz))

    def _handle_expr_Tmp(self, expr: Tmp):
        if self.use_tmps:
            codeloc = self._codeloc()
            self.def_to_loc.append((expr, codeloc))
            if codeloc not in self.loc_to_defs:
                self.loc_to_defs[codeloc] = OrderedSet()
            self.loc_to_defs[codeloc].add(expr)

            self.state.live_tmps.add(expr.tmp_idx)

    def _handle_binop_Default(self, expr: BinaryOp):
        self._expr(expr.operands[0])
        self._expr(expr.operands[1])

    _handle_binop_CmpLE = _handle_binop_Default
    _handle_binop_CmpLT = _handle_binop_Default
    _handle_binop_CmpGE = _handle_binop_Default
    _handle_binop_CmpGT = _handle_binop_Default
    _handle_binop_CmpEQ = _handle_binop_Default
    _handle_binop_CmpNE = _handle_binop_Default
    _handle_binop_Add = _handle_binop_Default
    _handle_binop_AddF = _handle_binop_Default
    _handle_binop_AddV = _handle_binop_Default
    _handle_binop_And = _handle_binop_Default
    _handle_binop_Carry = _handle_binop_Default
    _handle_binop_CmpF = _handle_binop_Default
    _handle_binop_Concat = _handle_binop_Default
    _handle_binop_Div = _handle_binop_Default
    _handle_binop_DivF = _handle_binop_Default
    _handle_binop_DivV = _handle_binop_Default
    _handle_binop_LogicalAnd = _handle_binop_Default
    _handle_binop_LogicalOr = _handle_binop_Default
    _handle_binop_Mod = _handle_binop_Default
    _handle_binop_Mul = _handle_binop_Default
    _handle_binop_Mull = _handle_binop_Default
    _handle_binop_MulF = _handle_binop_Default
    _handle_binop_MulV = _handle_binop_Default
    _handle_binop_MulHiV = _handle_binop_Default
    _handle_binop_Or = _handle_binop_Default
    _handle_binop_Rol = _handle_binop_Default
    _handle_binop_Ror = _handle_binop_Default
    _handle_binop_SBorrow = _handle_binop_Default
    _handle_binop_SCarry = _handle_binop_Default
    _handle_binop_Sar = _handle_binop_Default
    _handle_binop_Shl = _handle_binop_Default
    _handle_binop_Shr = _handle_binop_Default
    _handle_binop_Sub = _handle_binop_Default
    _handle_binop_SubF = _handle_binop_Default
    _handle_binop_SubV = _handle_binop_Default
    _handle_binop_Xor = _handle_binop_Default
    _handle_binop_InterleaveLOV = _handle_binop_Default
    _handle_binop_InterleaveHIV = _handle_binop_Default
    _handle_binop_CasCmpEQ = _handle_binop_Default
    _handle_binop_CasCmpNE = _handle_binop_Default
    _handle_binop_ExpCmpNE = _handle_binop_Default
    _handle_binop_SarNV = _handle_binop_Default
    _handle_binop_ShrNV = _handle_binop_Default
    _handle_binop_ShlNV = _handle_binop_Default
    _handle_binop_CmpEQV = _handle_binop_Default
    _handle_binop_CmpNEV = _handle_binop_Default
    _handle_binop_CmpGEV = _handle_binop_Default
    _handle_binop_CmpGTV = _handle_binop_Default
    _handle_binop_CmpLEV = _handle_binop_Default
    _handle_binop_CmpLTV = _handle_binop_Default
    _handle_binop_MinV = _handle_binop_Default
    _handle_binop_MaxV = _handle_binop_Default
    _handle_binop_QAddV = _handle_binop_Default
    _handle_binop_QNarrowBinV = _handle_binop_Default
    _handle_binop_PermV = _handle_binop_Default
    _handle_binop_Set = _handle_binop_Default

    def _handle_unop_Default(self, expr):
        self._expr(expr.operand)

    _handle_unop_BitwiseNeg = _handle_unop_Default
    _handle_unop_Dereference = _handle_unop_Default
    _handle_unop_Neg = _handle_unop_Default
    _handle_unop_Not = _handle_unop_Default
    _handle_unop_Reference = _handle_unop_Default
    _handle_unop_Clz = _handle_unop_Default
    _handle_unop_Ctz = _handle_unop_Default
    _handle_unop_GetMSBs = _handle_unop_Default
    _handle_unop_unpack = _handle_unop_Default
    _handle_unop_Sqrt = _handle_unop_Default
    _handle_unop_RSqrtEst = _handle_unop_Default

    def _handle_expr_UnaryOp(self, expr):
        self._expr(expr.operand)

    def _handle_expr_BinaryOp(self, expr):
        self._expr(expr.operands[0])
        self._expr(expr.operands[1])

    def _handle_expr_ITE(self, expr: ITE):
        self._expr(expr.cond)
        self._expr(expr.iftrue)
        self._expr(expr.iffalse)

    def _handle_expr_VEXCCallExpression(self, expr: VEXCCallExpression):
        for operand in expr.operands:
            self._expr(operand)

    def _handle_expr_DirtyExpression(self, expr: DirtyExpression):
        for operand in expr.operands:
            self._expr(operand)
        if expr.guard is not None:
            self._expr(expr.guard)
        if expr.maddr is not None:
            self._expr(expr.maddr)

    _handle_expr_Convert = _handle_unop_Default
    _handle_expr_Reinterpret = _handle_unop_Default

    def _handle_Dummy(self, expr):
        pass

    _handle_expr_VirtualVariable = _handle_Dummy
    _handle_expr_Phi = _handle_Dummy
    _handle_expr_Const = _handle_Dummy
    _handle_expr_MultiStatementExpression = _handle_Dummy
    _handle_expr_BasePointerOffset = _handle_Dummy
    _handle_expr_Call = _handle_Dummy



from __future__ import annotations
from angr.ailment.expression import ITE, Convert, BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class SingleBitCondToBoolExpr(PeepholeOptimizationExprBase):
    """
    Convert single-bit conditions to bool expressions
    """

    __slots__ = ()

    NAME = "Convert single-bit conditions to bool expressions (== 0 or == 1)"
    expr_classes = (ITE,)

    def optimize(self, expr: ITE, **kwargs):
        if isinstance(expr.cond, Convert) and expr.cond.to_bits == 1 and expr.cond.from_bits > 1:
            cond_inner = expr.cond.operand
        else:
            cond_inner = expr.cond

        if isinstance(cond_inner, BinaryOp):
            optimized = False
            if (
                cond_inner.op == "Xor"
                and isinstance(cond_inner.operands[1], Const)
                and cond_inner.operands[1].value == 1
            ):
                # A ^ 1 ==> !A, if A is single-bit
                # A ^ 1 ==> A == 0, if A is multi-bit
                optimized = True
                op0, op1 = cond_inner.operands
                if isinstance(op0, BinaryOp) and op0.op in BinaryOp.COMPARISON_NEGATION:
                    new_cond = BinaryOp(
                        None,
                        BinaryOp.COMPARISON_NEGATION[op0.op],
                        [op0.operands[0], op0.operands[1]],
                        False,
                        **op0.tags,
                    )
                else:
                    new_cond = BinaryOp(
                        None,
                        "CmpEQ",
                        [op0, Const(None, None, 0, op0.bits)],
                        False,
                        **cond_inner.tags,
                    )
            else:
                # A ==> A, if A is single-bit
                # A ==> A == 1, if A is multi-bit
                if cond_inner.op in BinaryOp.COMPARISON_NEGATION:
                    new_cond = cond_inner
                else:
                    op0, op1 = cond_inner.operands
                    if cond_inner.op == "CmpEQ" and isinstance(op1, Const) and op1.value == 1:
                        new_cond = cond_inner
                    else:
                        optimized = True
                        new_cond = BinaryOp(
                            None,
                            "CmpEQ",
                            [cond_inner, Const(None, None, 1, cond_inner.bits)],
                            False,
                            **cond_inner.tags,
                        )
            return (
                ITE(
                    expr.idx,
                    new_cond,
                    expr.iffalse,
                    expr.iftrue,
                    variable=expr.variable,
                    variable_offset=expr.variable_offset,
                    **expr.tags,
                )
                if optimized
                else None
            )

        return None



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("d2d1.dll")
prototypes = \
    {
        #
        'D2D1CreateFactory': SimTypeFunction([SimTypeInt(signed=False, label="D2D1_FACTORY_TYPE"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeRef("D2D1_FACTORY_OPTIONS", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["factoryType", "riid", "pFactoryOptions", "ppIFactory"]),
        #
        'D2D1MakeRotateMatrix': SimTypeFunction([SimTypeFloat(size=32), SimTypeRef("D2D_POINT_2F", SimStruct), SimTypePointer(SimTypeRef("D2D_MATRIX_3X2_F", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["angle", "center", "matrix"]),
        #
        'D2D1MakeSkewMatrix': SimTypeFunction([SimTypeFloat(size=32), SimTypeFloat(size=32), SimTypeRef("D2D_POINT_2F", SimStruct), SimTypePointer(SimTypeRef("D2D_MATRIX_3X2_F", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["angleX", "angleY", "center", "matrix"]),
        #
        'D2D1IsMatrixInvertible': SimTypeFunction([SimTypePointer(SimTypeRef("D2D_MATRIX_3X2_F", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["matrix"]),
        #
        'D2D1InvertMatrix': SimTypeFunction([SimTypePointer(SimTypeRef("D2D_MATRIX_3X2_F", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["matrix"]),
        #
        'D2D1CreateDevice': SimTypeFunction([SimTypeBottom(label="IDXGIDevice"), SimTypePointer(SimTypeRef("D2D1_CREATION_PROPERTIES", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="ID2D1Device"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dxgiDevice", "creationProperties", "d2dDevice"]),
        #
        'D2D1CreateDeviceContext': SimTypeFunction([SimTypeBottom(label="IDXGISurface"), SimTypePointer(SimTypeRef("D2D1_CREATION_PROPERTIES", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="ID2D1DeviceContext"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dxgiSurface", "creationProperties", "d2dDeviceContext"]),
        #
        'D2D1ConvertColorSpace': SimTypeFunction([SimTypeInt(signed=False, label="D2D1_COLOR_SPACE"), SimTypeInt(signed=False, label="D2D1_COLOR_SPACE"), SimTypePointer(SimTypeRef("D2D1_COLOR_F", SimStruct), offset=0)], SimTypeRef("D2D1_COLOR_F", SimStruct), arg_names=["sourceColorSpace", "destinationColorSpace", "color"]),
        #
        'D2D1SinCos': SimTypeFunction([SimTypeFloat(size=32), SimTypePointer(SimTypeFloat(size=32), offset=0), SimTypePointer(SimTypeFloat(size=32), offset=0)], SimTypeBottom(label="Void"), arg_names=["angle", "s", "c"]),
        #
        'D2D1Tan': SimTypeFunction([SimTypeFloat(size=32)], SimTypeFloat(size=32), arg_names=["angle"]),
        #
        'D2D1Vec3Length': SimTypeFunction([SimTypeFloat(size=32), SimTypeFloat(size=32), SimTypeFloat(size=32)], SimTypeFloat(size=32), arg_names=["x", "y", "z"]),
        #
        'D2D1ComputeMaximumScaleFactor': SimTypeFunction([SimTypePointer(SimTypeRef("D2D_MATRIX_3X2_F", SimStruct), offset=0)], SimTypeFloat(size=32), arg_names=["matrix"]),
        #
        'D2D1GetGradientMeshInteriorPointsFromCoonsPatch': SimTypeFunction([SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0), SimTypePointer(SimTypeRef("D2D_POINT_2F", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["pPoint0", "pPoint1", "pPoint2", "pPoint3", "pPoint4", "pPoint5", "pPoint6", "pPoint7", "pPoint8", "pPoint9", "pPoint10", "pPoint11", "pTensorPoint11", "pTensorPoint12", "pTensorPoint21", "pTensorPoint22"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations
from .flirt import FlirtAnalysis
from .flirt_sig import FlirtSignature, FlirtSignatureParsed, FlirtSignatureError
from .consts import FLIRT_ARCH_TO_ARCHNAME, FLIRT_OS_TO_OSNAME, FlirtAppType, FlirtOSType


def flirt_arch_to_arch_name(flirt_arch: int, app_types: int) -> str:
    """
    Convert FLIRT architecture ID to architecture name.

    :param flirt_arch: FLIRT architecture ID.
    :param app_types: FLIRT application types.
    :return: Architecture name.
    """
    try:
        arches = FLIRT_ARCH_TO_ARCHNAME[flirt_arch]
    except KeyError:
        return "Unknown"
    if app_types & FlirtAppType.APP_32_BIT and 32 in arches:
        return arches[32]
    if app_types & FlirtAppType.APP_64_BIT and 64 in arches:
        return arches[64]
    return "Unknown"


def flirt_os_type_to_os_name(os_type: int) -> str:
    """
    Convert FLIRT OS type to OS name.

    :param os_type: FLIRT OS type.
    :return: OS name.
    """
    try:
        v = FlirtOSType(os_type)
        return FLIRT_OS_TO_OSNAME.get(v, v.name)
    except ValueError:
        return "UnknownOS"


__all__ = [
    "FlirtAnalysis",
    "FlirtSignature",
    "FlirtSignatureError",
    "FlirtSignatureParsed",
    "flirt_arch_to_arch_name",
    "flirt_os_type_to_os_name",
]



from __future__ import annotations
import itertools
from typing import TYPE_CHECKING
import logging

from angr.analyses import Analysis, register_analysis
from angr.analyses.decompiler.condition_processor import ConditionProcessor
from angr.analyses.decompiler.graph_region import GraphRegion
from angr.analyses.decompiler.jumptable_entry_condition_rewriter import JumpTableEntryConditionRewriter
from angr.analyses.decompiler.empty_node_remover import EmptyNodeRemover
from angr.analyses.decompiler.jump_target_collector import JumpTargetCollector
from angr.analyses.decompiler.redundant_label_remover import RedundantLabelRemover
from angr.utils.graph import GraphUtils
from .structurer_nodes import BaseNode
from .structurer_base import StructurerBase
from .dream import DreamStructurer


if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function


_l = logging.getLogger(__name__)


class RecursiveStructurer(Analysis):
    """
    Recursively structure a region and all of its subregions.
    """

    def __init__(
        self,
        region,
        cond_proc=None,
        func: Function | None = None,
        structurer_cls: type | None = None,
        **kwargs,
    ):
        self._region = region
        self.cond_proc = cond_proc if cond_proc is not None else ConditionProcessor(self.project.arch)
        self.function = func
        self.structurer_cls = structurer_cls if structurer_cls is not None else DreamStructurer
        self.structurer_options = kwargs

        self.result: BaseNode | None = None
        self.result_incomplete: bool = False

        self._analyze()

    def _analyze(self):
        region = self._region.recursive_copy()
        self._case_entry_to_switch_head: dict[int, int] = self._get_switch_case_entries()
        self.result_incomplete = False

        # visit the region in post-order DFS
        parent_map = {}
        stack = [region]

        while stack:
            current_region = stack[-1]

            has_region = False
            for node in GraphUtils.dfs_postorder_nodes_deterministic(current_region.graph, current_region.head):
                subnodes = []
                if type(node) is GraphRegion:
                    if node.cyclic:
                        subnodes.append(node)
                    else:
                        subnodes.insert(0, node)
                    parent_map[node] = current_region
                    has_region = True
                # remove existing regions
                for subnode in subnodes:
                    if subnode in stack:
                        stack.remove(subnode)
                stack.extend(subnodes)

            if not has_region:
                # pop this region from the stack
                stack.pop()

                # Get the parent region
                parent_region = parent_map.get(current_region)
                # structure this region
                st: StructurerBase = self.project.analyses[self.structurer_cls].prep(
                    kb=self.kb, fail_fast=self._fail_fast
                )(
                    current_region.copy(),
                    parent_map=parent_map,
                    condition_processor=self.cond_proc,
                    case_entry_to_switch_head=self._case_entry_to_switch_head,
                    func=self.function,
                    parent_region=parent_region,
                    jump_tables=self.kb.cfgs["CFGFast"].jump_tables,
                    **self.structurer_options,
                )
                # replace this region with the resulting node in its parent region... if it's not an orphan
                if not parent_region:
                    # this is the top-level region. we are done!
                    if st.result is None:
                        # take the partial result out of the graph
                        _l.warning(
                            "Structuring failed to complete (most likely due to bugs in structuring). The "
                            "output will miss code blocks."
                        )
                        self.result = self._pick_incomplete_result_from_region(st._region)
                        self.result_incomplete = True
                    else:
                        self.result = st.result
                    break

                if st.result is None:
                    self._replace_region_with_region(parent_region, current_region, st._region)
                else:
                    self._replace_region_with_node(
                        parent_region, current_region, st._region, st.result, st.virtualized_edges
                    )

        if self.structurer_cls is DreamStructurer:
            # rewrite conditions in the result to remove all jump table entry conditions
            rewriter = JumpTableEntryConditionRewriter(set(itertools.chain(*self.cond_proc.jump_table_conds.values())))
            rewriter.walk(self.result)  # update SequenceNodes in-place

            # remove all goto statements
            # TODO: Properly implement support for multi-entry regions
            StructurerBase._remove_all_jumps(self.result)

        else:
            StructurerBase._remove_redundant_jumps(self.result)

        # remove redundant labels
        jtc = JumpTargetCollector(self.result)
        self.result = RedundantLabelRemover(self.result, jtc.jump_targets).result

        # remove empty nodes (if any)
        self.result = EmptyNodeRemover(self.result).result

        if self.structurer_cls is DreamStructurer:
            # remove conditional jumps
            StructurerBase._remove_conditional_jumps(self.result)

        self.result = self.cond_proc.remove_claripy_bool_asts(self.result)

    @staticmethod
    def _replace_region_with_node(parent_region, sub_region, updated_sub_region, node, virtualized_edges):
        parent_region.replace_region(sub_region, updated_sub_region, node, virtualized_edges)

    @staticmethod
    def _replace_region_with_region(parent_region, sub_region, new_region):
        parent_region.replace_region_with_region(sub_region, new_region)

    def _get_switch_case_entries(self) -> dict[int, int]:
        if self.function is None:
            return {}

        entries = {}
        func_block_addrs = self.function.block_addrs_set

        jump_tables = self.kb.cfgs["CFGFast"].jump_tables
        for jump_table_head_addr, jumptable in jump_tables.items():
            if jump_table_head_addr not in func_block_addrs:
                continue
            assert jumptable.jumptable_entries is not None
            for entry_addr in jumptable.jumptable_entries:
                entries[entry_addr] = jump_table_head_addr

        return entries

    def _pick_incomplete_result_from_region(self, region):
        """
        Parse the region graph and get (a) the node with address equal to the function address, or (b) the node with
        the lowest address.
        """

        min_node = None
        for node in region.graph.nodes:
            if not isinstance(node, BaseNode):
                continue
            if self.function is not None and node.addr == self.function.addr:
                return node
            if min_node is None or (min_node.addr is not None and node.addr is not None and min_node.addr < node.addr):
                min_node = node

        return min_node


register_analysis(RecursiveStructurer, "RecursiveStructurer")



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("mspatcha.dll")
prototypes = \
    {
        #
        'TestApplyPatchToFileA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileName", "OldFileName", "ApplyOptionFlags"]),
        #
        'TestApplyPatchToFileW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileName", "OldFileName", "ApplyOptionFlags"]),
        #
        'TestApplyPatchToFileByHandles': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileHandle", "OldFileHandle", "ApplyOptionFlags"]),
        #
        'TestApplyPatchToFileByBuffers': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileBuffer", "PatchFileSize", "OldFileBuffer", "OldFileSize", "NewFileSize", "ApplyOptionFlags"]),
        #
        'ApplyPatchToFileA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileName", "OldFileName", "NewFileName", "ApplyOptionFlags"]),
        #
        'ApplyPatchToFileW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileName", "OldFileName", "NewFileName", "ApplyOptionFlags"]),
        #
        'ApplyPatchToFileByHandles': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileHandle", "OldFileHandle", "NewFileHandle", "ApplyOptionFlags"]),
        #
        'ApplyPatchToFileExA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["CallbackContext", "CurrentPosition", "MaximumPosition"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileName", "OldFileName", "NewFileName", "ApplyOptionFlags", "ProgressCallback", "CallbackContext"]),
        #
        'ApplyPatchToFileExW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["CallbackContext", "CurrentPosition", "MaximumPosition"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileName", "OldFileName", "NewFileName", "ApplyOptionFlags", "ProgressCallback", "CallbackContext"]),
        #
        'ApplyPatchToFileByHandlesEx': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["CallbackContext", "CurrentPosition", "MaximumPosition"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileHandle", "OldFileHandle", "NewFileHandle", "ApplyOptionFlags", "ProgressCallback", "CallbackContext"]),
        #
        'ApplyPatchToFileByBuffers': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("FILETIME", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["CallbackContext", "CurrentPosition", "MaximumPosition"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["PatchFileMapped", "PatchFileSize", "OldFileMapped", "OldFileSize", "NewFileBuffer", "NewFileBufferSize", "NewFileActualSize", "NewFileTime", "ApplyOptionFlags", "ProgressCallback", "CallbackContext"]),
        #
        'GetFilePatchSignatureA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_IGNORE_RANGE", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_RETAIN_RANGE", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["FileName", "OptionFlags", "OptionData", "IgnoreRangeCount", "IgnoreRangeArray", "RetainRangeCount", "RetainRangeArray", "SignatureBufferSize", "SignatureBuffer"]),
        #
        'GetFilePatchSignatureW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_IGNORE_RANGE", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_RETAIN_RANGE", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["FileName", "OptionFlags", "OptionData", "IgnoreRangeCount", "IgnoreRangeArray", "RetainRangeCount", "RetainRangeArray", "SignatureBufferSize", "SignatureBuffer"]),
        #
        'GetFilePatchSignatureByHandle': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_IGNORE_RANGE", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_RETAIN_RANGE", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["FileHandle", "OptionFlags", "OptionData", "IgnoreRangeCount", "IgnoreRangeArray", "RetainRangeCount", "RetainRangeArray", "SignatureBufferSize", "SignatureBuffer"]),
        #
        'GetFilePatchSignatureByBuffer': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_IGNORE_RANGE", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_RETAIN_RANGE", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["FileBufferWritable", "FileSize", "OptionFlags", "OptionData", "IgnoreRangeCount", "IgnoreRangeArray", "RetainRangeCount", "RetainRangeArray", "SignatureBufferSize", "SignatureBuffer"]),
        #
        'NormalizeFileForPatchSignature': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_OPTION_DATA", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_IGNORE_RANGE", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PATCH_RETAIN_RANGE", SimStruct), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["FileBuffer", "FileSize", "OptionFlags", "OptionData", "NewFileCoffBase", "NewFileCoffTime", "IgnoreRangeCount", "IgnoreRangeArray", "RetainRangeCount", "RetainRangeArray"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("mssign32.dll")
prototypes = \
    {
        #
        'SignError': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'SignerFreeSignerContext': SimTypeFunction([SimTypePointer(SimTypeRef("SIGNER_CONTEXT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSignerContext"]),
        #
        'SignerSign': SimTypeFunction([SimTypePointer(SimTypeRef("SIGNER_SUBJECT_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_CERT", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_SIGNATURE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_PROVIDER_INFO", SimStruct), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("CRYPT_ATTRIBUTES", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSubjectInfo", "pSignerCert", "pSignatureInfo", "pProviderInfo", "pwszHttpTimeStamp", "psRequest", "pSipData"]),
        #
        'SignerSignEx': SimTypeFunction([SimTypeInt(signed=False, label="SIGNER_SIGN_FLAGS"), SimTypePointer(SimTypeRef("SIGNER_SUBJECT_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_CERT", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_SIGNATURE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_PROVIDER_INFO", SimStruct), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("CRYPT_ATTRIBUTES", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("SIGNER_CONTEXT", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwFlags", "pSubjectInfo", "pSignerCert", "pSignatureInfo", "pProviderInfo", "pwszHttpTimeStamp", "psRequest", "pSipData", "ppSignerContext"]),
        #
        'SignerSignEx2': SimTypeFunction([SimTypeInt(signed=False, label="SIGNER_SIGN_FLAGS"), SimTypePointer(SimTypeRef("SIGNER_SUBJECT_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_CERT", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_SIGNATURE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_PROVIDER_INFO", SimStruct), offset=0), SimTypeInt(signed=False, label="SIGNER_TIMESTAMP_FLAGS"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("CRYPT_ATTRIBUTES", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("SIGNER_CONTEXT", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeRef("CERT_STRONG_SIGN_PARA", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwFlags", "pSubjectInfo", "pSignerCert", "pSignatureInfo", "pProviderInfo", "dwTimestampFlags", "pszTimestampAlgorithmOid", "pwszHttpTimeStamp", "psRequest", "pSipData", "ppSignerContext", "pCryptoPolicy", "pReserved"]),
        #
        'SignerSignEx3': SimTypeFunction([SimTypeInt(signed=False, label="SIGNER_SIGN_FLAGS"), SimTypePointer(SimTypeRef("SIGNER_SUBJECT_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_CERT", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_SIGNATURE_INFO", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_PROVIDER_INFO", SimStruct), offset=0), SimTypeInt(signed=False, label="SIGNER_TIMESTAMP_FLAGS"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("CRYPT_ATTRIBUTES", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("SIGNER_CONTEXT", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeRef("CERT_STRONG_SIGN_PARA", SimStruct), offset=0), SimTypePointer(SimTypeRef("SIGNER_DIGEST_SIGN_INFO", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwFlags", "pSubjectInfo", "pSignerCert", "pSignatureInfo", "pProviderInfo", "dwTimestampFlags", "pszTimestampAlgorithmOid", "pwszHttpTimeStamp", "psRequest", "pSipData", "ppSignerContext", "pCryptoPolicy", "pDigestSignInfo", "pReserved"]),
        #
        'SignerTimeStamp': SimTypeFunction([SimTypePointer(SimTypeRef("SIGNER_SUBJECT_INFO", SimStruct), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("CRYPT_ATTRIBUTES", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSubjectInfo", "pwszHttpTimeStamp", "psRequest", "pSipData"]),
        #
        'SignerTimeStampEx': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("SIGNER_SUBJECT_INFO", SimStruct), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("CRYPT_ATTRIBUTES", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("SIGNER_CONTEXT", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwFlags", "pSubjectInfo", "pwszHttpTimeStamp", "psRequest", "pSipData", "ppSignerContext"]),
        #
        'SignerTimeStampEx2': SimTypeFunction([SimTypeInt(signed=False, label="SIGNER_TIMESTAMP_FLAGS"), SimTypePointer(SimTypeRef("SIGNER_SUBJECT_INFO", SimStruct), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="ALG_ID"), SimTypePointer(SimTypeRef("CRYPT_ATTRIBUTES", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("SIGNER_CONTEXT", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwFlags", "pSubjectInfo", "pwszHttpTimeStamp", "dwAlgId", "psRequest", "pSipData", "ppSignerContext"]),
        #
        'SignerTimeStampEx3': SimTypeFunction([SimTypeInt(signed=False, label="SIGNER_TIMESTAMP_FLAGS"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("SIGNER_SUBJECT_INFO", SimStruct), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("CRYPT_ATTRIBUTES", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("SIGNER_CONTEXT", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeRef("CERT_STRONG_SIGN_PARA", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwFlags", "dwIndex", "pSubjectInfo", "pwszHttpTimeStamp", "pszAlgorithmOid", "psRequest", "pSipData", "ppSignerContext", "pCryptoPolicy", "pReserved"]),
    }

lib.set_prototypes(prototypes)
